#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "clm.h"

clm_Matrix INVALID_MATRIX = { .rows = 0, .cols = 0, .values = NULL };

clm_Vector clm_createVector(int length) {
	clm_Vector vector;
	vector.values = calloc(length, sizeof(float));
	vector.length = length;
	return vector;
}

void clm_freeVector(clm_Vector vector) {
	free(vector.values);
}

clm_Layer clm_createLayer(unsigned int prevNumNeurons, unsigned int numLayers) {
	clm_Layer layer;
	for(unsigned int i = 0; i < numLayers; i++) {

	}
	return layer;
}

clm_NN clm_createNN(unsigned int numLayers, unsigned int *layerSizes) {
	clm_NN nn;
	nn.numLayers = numLayers;
	for(unsigned int i = 0; i < numLayers; i++) {
		clm_Layer layer;

	}
	return nn;
}

clm_Matrix clm_createMatrix(unsigned int rows, unsigned int cols) {
	clm_Matrix mat;
	mat.rows = rows;
	mat.cols = cols;
	mat.values = calloc(rows * cols, sizeof(float));
	return mat;
}

clm_Matrix clm_createMatrixRandom(unsigned int rows, unsigned int cols) {
	clm_Matrix mat = clm_createMatrix(rows, cols);

	for(unsigned int i = 0; i < rows; i++) {
		for(unsigned int j = 0; j < cols; j++) {
			mat.values[i * cols + j] = ((float) rand() / RAND_MAX) * 2.0f - 1.0f;
		}
	}

	return mat;
}

void clm_freeMatrix(clm_Matrix mat) {
	free(mat.values);
}

clm_Matrix clm_copyMatrix(clm_Matrix mat) {
	clm_Matrix copy = clm_createMatrix(mat.rows, mat.cols);

	memcpy(copy.values, mat.values, mat.rows * mat.cols * sizeof(float));

	return copy;
}

clm_Matrix clm_matrixAddScalar(clm_Matrix mat, float scalar) {
	for(unsigned int i = 0; i < mat.cols * mat.rows; i++) {
		mat.values[i] += scalar;
	}

	return mat;
}

clm_Matrix clm_matrixAddMatrix(clm_Matrix mat, clm_Matrix other) {
	if(mat.cols != other.cols || mat.rows != other.rows) {
		printf("Failed to add matrices\n");
		return INVALID_MATRIX;
	}

	for(unsigned int i = 0; i < mat.cols * mat.rows; i++) {
		mat.values[i] += other.values[i];
	}

	clm_freeMatrix(other);

	return mat;
}

clm_Matrix clm_matrixSubtractMatrix(clm_Matrix mat, clm_Matrix other) {
	if(mat.cols != other.cols || mat.rows != other.rows) {
		printf("Failed to add matrices\n");
		return INVALID_MATRIX;
	}

	for(unsigned int i = 0; i < mat.rows * mat.cols; i++) {
		mat.values[i] -= other.values[i];
	}

	clm_freeMatrix(other);

	return mat;
}

clm_Matrix clm_matrixTranspose(clm_Matrix mat) {
	clm_Matrix tr = clm_createMatrix(mat.cols, mat.rows);
	for(unsigned int i = 0; i < tr.rows; i++) {
		for(unsigned int j = 0; j < tr.cols; j++) {
			tr.values[i * tr.cols + j] = mat.values[j * mat.cols + i];
		}
	}

	clm_freeMatrix(mat);

	return tr;
}

clm_Matrix clm_matrixMultiplyMatrix(clm_Matrix a, clm_Matrix b) {
	if(a.cols != b.rows) {
		printf("Cannot multiply matrices\n");
		return INVALID_MATRIX;
	}

	clm_Matrix r = clm_createMatrix(a.rows, b.cols);
	for(unsigned int i = 0; i < r.rows; i++) {
		for(unsigned int j = 0; j < r.cols; j++) {
			float sum = 0;
			for(unsigned int k = 0; k < a.cols; k++) {
				sum += a.values[i * a.cols + k] * b.values[k * b.cols + j];
			}
			r.values[i * r.cols + j] = sum;
		}
	}

	clm_freeMatrix(a);
	clm_freeMatrix(b);

	return r;
}

clm_Matrix clm_matrixMultiplyScalar(clm_Matrix mat, float scalar) {
	for(unsigned int i = 0; i < mat.cols * mat.rows; i++) {
		mat.values[i] *= scalar;
	}

	return mat;
}

clm_Matrix clm_matrixSigmoid(clm_Matrix mat) {
	for(unsigned int i = 0; i < mat.rows; i++) {
		for(unsigned int j = 0; j < mat.cols; j++) {
			mat.data[i * mat.cols + j] = 1 / (1 + exp(-mat.values[i * mat.cols + j]));
		}
	}

	return mat;
}

clm_Matrix clm_matrixDSigmoid(clm_Matrix mat) {
	for(unsigned int i = 0; i < mat.rows; i++) {
		for(unsigned int j = 0; j < mat.cols; j++) {
			float v = mat.data[i * mat.cols + j];
			mat.data[i * mat.cols + j] = v * (1 - v);
		}
	}

	return mat;
}

void clm_matrixPrint(clm_Matrix mat) {
	printf("[\n");
	for(unsigned int i = 0; i < mat.rows; i++) {
		for(unsigned int j = 0; j < mat.cols; j++) {
			printf("%7.3f", mat.values[i * mat.cols + j]);
		}
		printf("\n");
	}
	printf("]\n");
}
